(*  Title:      ASP.thy
    ID:         $Id: FinASP.thy,v 1.7 2007/03/05 18:42:06 hsudhof Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2006

    Note:       Based on a formalization of Abadi/Cardelli's
                Sigma calculus in ZF by Sidi O'Ehmety
*)

header {* Basic definitions of Sigma-calculus enriched with activities 
          and future references *}

theory FinASP imports FMap begin

subsection {* Object-terms in de Bruijn notation and substitution *}

(* types Label = nat   *)
consts max_label :: nat

typedef Label = "{n :: nat. n <= max_label }"
by auto;


lemma finite_Label_set: "finite {n :: nat. n <= max_label }";
thm nat_seg_image_imp_finite;
apply (rule_tac n = "max_label + 1" and f = "% x. x" in nat_seg_image_imp_finite);
apply (simp add: ran_def);
apply force;
done;

lemma Univ_abs_label: "(UNIV :: (Label set)) = Abs_Label ` {n :: nat. n <= max_label }";
apply (subgoal_tac "! x :: Label. x : Abs_Label ` {n :: nat. n <= max_label }");
apply force;
thm imageI;
apply (rule allI);
apply (subgoal_tac "Abs_Label(Rep_Label x) = x");
prefer 2;
apply (rule Rep_Label_inverse) ;
apply (erule subst);
apply (rule imageI);
apply (fold Label_def);
apply (rule Rep_Label);
done;



lemma finite_Label: "finite (UNIV :: (Label set))";
apply (subst Univ_abs_label);
apply (rule finite_imageI);
apply (rule finite_Label_set);
done;

instance Label :: "fintype" 
proof 
show "finite (UNIV :: (Label set))" by (rule finite_Label);
qed;

consts
Lsuc :: "(Label set) \<Rightarrow> Label \<Rightarrow> Label"
Lmin :: "(Label set) \<Rightarrow> Label"
Lmax :: "(Label set) \<Rightarrow> Label"

constdefs 
Llt :: "[Label,  Label] \<Rightarrow> bool"    (infixl "<" 50)
"Llt a b == Rep_Label a < Rep_Label b"
Lle :: "[Label,  Label] \<Rightarrow> bool"    (infixl "\<le>" 50)
"Lle a b == Rep_Label a \<le> Rep_Label b"

datatype ActivityRef = aref nat

datatype FutureRef = fref nat

datatype type =  Object "Label -~> (type \<times> type)"

datatype Variable = Self nat | Param nat 
(* each binder introduces 2 variables: self and parameter *)
(* thus to each debruijn index (nat) correspond 2 vartiables of the same depth *)

datatype dB = (* the typed ASPfin datatype *) 
    Var Variable              (* variable -- as deBruijn index *)
  | Obj "Label -~>dB" type    (* An object maps labels to terms and has a type *)
  | Call dB Label dB          (* Call a l b calls method l on object a with param b *)
  | Upd dB Label dB           (* Upd a l b updates method l of a  with body b *)
  | Active dB                 (* creates an active object *)
  | ActRef ActivityRef        (* references an active object -- dynamic syntax *)
  | FutRef FutureRef          (* references a future -- dynamic syntax *)

consts
  lift ::         "[dB, nat] \<Rightarrow> dB"
  lift_Var ::     "[Variable,nat] \<Rightarrow> Variable"
  lift_option ::  "[nat, dB option] \<Rightarrow> dB option"
  subst ::        "[dB, dB, dB, nat] \<Rightarrow> dB"  ("_[_,_'/_]" [300, 0, 0,0] 300)
  subst_Var ::    "[Variable, dB, dB, nat] \<Rightarrow> dB" 
  subst_option :: "[nat, dB, dB, dB option] \<Rightarrow> dB option"  


(* lifting and substition for de Bruijn terms *)
primrec
  lift_Self:  "lift_Var (Self i) k = Self (if i < k then i else (i + 1))"
  lift_Param: "lift_Var (Param i) k = Param (if i < k then i else (i + 1))"
primrec
  liftVar:    "lift (Var v) k = Var (lift_Var v k) "
  liftCall:   "lift (Call a l b) k = Call (lift a k) l (lift b k)"
  liftUpd:    "lift (Upd a l b) k = Upd (lift a k) l (lift b (k + 1))" 
  liftObj:    "lift (Obj f T) k = Obj (% l. (lift_option (k + 1) (f l))) T" 
  lift_None:  "lift_option k None = None"
  lift_Some:  "lift_option k (Some t) = Some (lift t k)"
  lift_Act:   "lift (Active a) k = Active (lift a k)"
  lift_ARef:  "lift (ActRef g) k = ActRef g"
  lift_FRef:  "lift (FutRef f) k = FutRef f"



primrec
  subst_Self:  "subst_Var (Self i) a b k = (if k < i then Var (Self (i - 1)) else if i = k then a else Var (Self i))"
  subst_Param: "subst_Var (Param i) a b k = (if k < i then Var (Param (i - 1)) else if i = k then b else Var (Param i))"
primrec  
  subst_Var:   "(Var i)[s,t/k] =subst_Var i s t k"
  subst_Call:  "(Call a l b)[s,t/k] = (Call (a [s,t/k]) l (b[s,t/k]))"
  subst_Upd:   "Upd a l b [s,t/k] = (Upd (a [s,t/k]) l (b [lift s 0,lift t 0 / k+1]))"
  subst_Obj:   "Obj f T [s,t/k] = Obj (% l. (subst_option (k+1)(lift s 0)(lift t 0)(f l))) T"
  subst_None:  "subst_option n s t None = None"
  subst_Some:  "subst_option n s t (Some a) = Some (a [s,t/n])"
  subst_Act:   "(Active a)[s,t/k] = Active (a [s,t/k])"
  subst_ARef:  "(ActRef g)[s,t/k] = ActRef g"
  subst_FRef:  "(FutRef f)[s,t/k] = FutRef f"


declare subst_Var [simp del]

subsection {* Local semantics *}

(* We left out any definedness assumptions in beta and upd -- these can be modelled
   easily (l > length f), but we first want to see if we need them at all *)

inductive beta :: "[dB, dB] => bool"  (infixl "\<rightarrow>\<^sub>\<beta>" 50)
  where
    beta [simp, intro!]: "l : dom f \<Longrightarrow> Call (Obj f T) l b \<rightarrow>\<^sub>\<beta>  (the(f l))[(Obj f T),b/0]" | 
           (* performs a method invocation *)
    upd  [simp, intro!]: "l : dom f \<Longrightarrow>
                         Upd (Obj f T) l a \<rightarrow>\<^sub>\<beta>  Obj (f (l \<mapsto> a)) T" |
           (* updates a method; creating a new object *)

     (* then other rules allow reduction inside contexts *)
    sel  [simp, intro!]: "s \<rightarrow>\<^sub>\<beta> t \<Longrightarrow> Call s l u \<rightarrow>\<^sub>\<beta>  Call t l u" |
    selR [simp, intro!]: "u \<rightarrow>\<^sub>\<beta> t \<Longrightarrow> Call s l u \<rightarrow>\<^sub>\<beta>  Call s l t" |
    updL [simp, intro!]: "s \<rightarrow>\<^sub>\<beta> t \<Longrightarrow> Upd s l u \<rightarrow>\<^sub>\<beta> Upd t l u" |
    updR [simp, intro!]: "s \<rightarrow>\<^sub>\<beta> t \<Longrightarrow> Upd u l s \<rightarrow>\<^sub>\<beta> Upd u l t" |
    obj [simp, intro!]: "\<lbrakk> s \<rightarrow>\<^sub>\<beta> t; l: dom f \<rbrakk> \<Longrightarrow> Obj (f (l \<mapsto> s)) T \<rightarrow>\<^sub>\<beta> Obj (f (l \<mapsto> t)) T" |
    act [simp, intro!]: "s \<rightarrow>\<^sub>\<beta> t \<Longrightarrow> Active s \<rightarrow>\<^sub>\<beta> Active t"    

abbreviation
  beta_reds :: "[dB, dB] => bool"  (infixl "->>" 50) where
  "s ->> t == beta^** s t"
(* NB an alternative semantics is given in ContextLocalRed.thy *)

notation (latex)
  beta_reds  (infixl "\<rightarrow>\<^sub>\<beta>\<^sup>*" 50)

inductive_cases beta_cases [elim!]:
  "Var i \<rightarrow>\<^sub>\<beta> t"
  "Call s l t\<rightarrow>\<^sub>\<beta> u"
  "Upd s l t \<rightarrow>\<^sub>\<beta> u"
  "Obj s T \<rightarrow>\<^sub>\<beta> Obj t T"

declare if_not_P [simp] not_less_eq [simp]
  -- {* don't add @{text "r_into_rtrancl[intro!]"} *}

subsection {* Congruence rules *}

lemma rtrancl_beta_ObjL [intro!]:
    "s \<rightarrow>\<^sub>\<beta>\<^sup>* s' ==> Call s l u\<rightarrow>\<^sub>\<beta>\<^sup>* Call s' l u"
  apply (erule rtranclp_induct)
  apply (blast intro: rtranclp.rtrancl_into_rtrancl)+
  done

lemma rtrancl_beta_ObjR [intro!]:
    "s \<rightarrow>\<^sub>\<beta>\<^sup>* s' ==> Call t l s \<rightarrow>\<^sub>\<beta>\<^sup>* Call t l s'"
  apply (erule rtranclp_induct)
  apply (blast intro: rtranclp.rtrancl_into_rtrancl)+
  done
  
lemma rtrancl_beta_Obj [intro!]:
    "\<lbrakk>s \<rightarrow>\<^sub>\<beta>\<^sup>* s';t\<rightarrow>\<^sub>\<beta>\<^sup>*t' \<rbrakk> \<Longrightarrow> Call s l t\<rightarrow>\<^sub>\<beta>\<^sup>* Call s' l t'";
  apply (erule rtranclp_induct)
  apply (blast intro: rtranclp.rtrancl_into_rtrancl)+
  done

lemma rtrancl_beta_updL:
"s \<rightarrow>\<^sub>\<beta>\<^sup>* s' \<Longrightarrow> Upd s l u \<rightarrow>\<^sub>\<beta>\<^sup>* Upd s' l u"
  apply (erule rtranclp_induct)
  apply (blast intro: rtranclp.rtrancl_into_rtrancl)+
  done

lemma rtrancl_beta_updR:
   "s \<rightarrow>\<^sub>\<beta>\<^sup>* s' \<Longrightarrow> Upd u l s \<rightarrow>\<^sub>\<beta>\<^sup>* Upd u l s'"
  apply (erule rtranclp_induct)
  apply (blast intro: rtranclp.rtrancl_into_rtrancl)+
  done

lemma rtrancl_beta_upd:
   "\<lbrakk> u \<rightarrow>\<^sub>\<beta>\<^sup>* u'; s \<rightarrow>\<^sub>\<beta>\<^sup>* s' \<rbrakk> \<Longrightarrow> Upd u l s \<rightarrow>\<^sub>\<beta>\<^sup>* Upd u' l s'"
  apply (blast intro!: rtrancl_beta_updL rtrancl_beta_updR
    intro: rtranclp_trans)
  done;

lemma rtrancl_beta_Act: "s \<rightarrow>\<^sub>\<beta>\<^sup>* s' \<Longrightarrow> Active s \<rightarrow>\<^sub>\<beta>\<^sup>* Active s'"
  apply (erule rtranclp_induct)
  apply (blast intro: rtranclp.rtrancl_into_rtrancl)+
  done

lemma rtrancl_beta_obj: 
  "\<lbrakk> s \<rightarrow>\<^sub>\<beta>\<^sup>* s'; l : dom f \<rbrakk> \<Longrightarrow> Obj (f (l \<mapsto> s)) T \<rightarrow>\<^sub>\<beta>\<^sup>* Obj (f (l \<mapsto> s')) T"
  apply (erule rtranclp_induct)
  apply(simp)
  apply (erule rtranclp.rtrancl_into_rtrancl)+
  apply(rule beta.intros)
by assumption+;

lemma obj_lem: "\<lbrakk> n: dom f; the (f n) ->> x\<rbrakk> \<Longrightarrow> Obj (f) T ->> Obj (f(n \<mapsto> x)) T"
  apply (rule_tac P = "% y. Obj y T ->> Obj (f (n \<mapsto> x)) T" and s = "(f (n \<mapsto> the (f n)))" in subst);
  apply (rule fun_upd_idem)
  apply (force)
  apply (erule rtrancl_beta_obj)
  apply (assumption)
done


constdefs
Ltake_eq ::" [Label set, (Label ~=> 'a),  (Label ~=> 'a)] \<Rightarrow> bool "
"Ltake_eq L f g  == \<forall>l \<in> L . f l = g l"

lemma empty_dom : " \<lbrakk>{} = dom g\<rbrakk> \<Longrightarrow> g = empty"
  apply(rule ext)
  apply(subgoal_tac "uu \<notin> dom g")
  apply (unfold dom_def)
  apply(fastsimp)
  apply (erule subst)
  apply (rule notI)
  apply(erule emptyE)
done

lemma Ltake_eq_all : "\<lbrakk>dom f = dom g; Ltake_eq (dom f) f g\<rbrakk> \<Longrightarrow> f = g"
  apply(unfold Ltake_eq_def)
  apply(subgoal_tac "\<forall>l. f l = g l" )
  apply(auto)
  apply(subgoal_tac "f \<subseteq>\<^sub>m g")
  apply(subgoal_tac "g \<subseteq>\<^sub>m f")
  apply(rule map_le_antisym)
  apply(assumption)+
  apply (unfold map_le_def)
  apply(erule subst)
  apply(simp)+
  apply(case_tac "l\<in> dom g")
  apply auto
done

lemma Ltake_eq_dom : "\<lbrakk> L \<subseteq> dom (f::(Label -~> 'a)) ; card L = card (dom f)\<rbrakk> \<Longrightarrow> L = (dom f)"
  apply(subgoal_tac "\<forall> l \<in> L . l \<in> (dom f)")
  apply(auto)
  apply (subgoal_tac "finite (dom f)")
  apply(drule_tac A=L in card_seteq)
  apply(simp)+
  apply(force)
  apply(rule  finite_dom_fmap)
done

lemma Ltake_eq_empty [simp]: "Ltake_eq {} f g"
  apply(unfold Ltake_eq_def)
  apply(simp)
done

 
 
lemma rtrancl_beta_obj_lem00: 
        "\<lbrakk> dom f = dom g; \<forall> l \<in>  dom  f. the (f l) ->> the (g l) \<rbrakk> \<Longrightarrow>  
        \<forall>  k  \<le> (card (dom f)) . 
          (\<exists> ob . length ob = (k + 1) \<and> 
          (\<forall>  obi. obi mem ob \<longrightarrow> dom (fst(obi)) =  dom f \<and> ((snd obi)  \<subseteq> dom f)) \<and> 
          (fst (ob ! 0) = f)  \<and>
          (card (snd (ob!k)) = k) \<and>
          (\<forall> i < k . snd (ob ! i ) \<subset> snd (ob ! k)) \<and> 
          (Obj (fst (ob ! 0)) T ->> Obj (fst (ob ! k)) T) \<and> 
          (card (snd (ob !k ))= k \<longrightarrow> (Ltake_eq (snd (ob!k)) (fst (ob ! k)) g) \<and>
                                       (Ltake_eq ((dom f) - (snd (ob!k))) (fst (ob ! k)) f)
          ))"
     
apply (rule allI);
apply (induct_tac k);
apply simp;
apply (rule_tac x = "[(f,{})]" in exI);
apply simp;
apply(erule subst)
apply(unfold Ltake_eq_def)
apply(simp)

apply (rule impI);
apply (subgoal_tac "n <= (card (dom f))");
defer

apply arith;
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule conjE)+;

apply(subgoal_tac "  n < card (dom f) \<longrightarrow> ( \<exists> l' . l' \<in> dom f \<and>  l' \<notin> snd (ob!(n))) ")

apply(rotate_tac -1)
apply(erule impE)
apply(arith)
apply(erule exE)

apply(rule_tac x="ob @[((fst (ob ! n))(l'\<mapsto> the  (g l')), insert l' (snd (ob ! n)) )] " in exI)
apply(rule conjI)
apply(simp)
apply(rule conjI)
apply(rule allI)
apply(rule impI)
apply(subgoal_tac 
        "obi =  (fst (ob ! n)(l' \<mapsto> the (g l')), insert l' (snd (ob ! n))) | obi mem ob")
apply(erule disjE)
apply(erule conjE)
apply(rule conjI)
apply(rotate_tac -3)
apply(erule_tac x="ob!n" in  allE)
apply(erule impE)
apply(simp)
apply(erule conjE)

apply(subgoal_tac "dom (fst obi) = dom  ((fst(ob ! n))(l' \<mapsto> the (g l')))")
apply(subgoal_tac "  dom (fst obi) = dom (fst (ob ! n)(l' \<mapsto> the (g l'))) ")
apply(rotate_tac -1)
apply(erule ssubst)
apply(simp)
apply(rotate_tac 2)
apply(unfold insert_def)

apply(force)
apply(simp)
apply(simp)
apply(subgoal_tac "snd obi = {x. x = l'} \<union> snd (ob ! n)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(erule_tac x="ob ! n" in allE)

apply(simp)
apply(simp)
apply(rule conjI)
apply(erule_tac x="obi" in allE)
apply(erule impE)
apply(assumption)
apply(erule conjE)
apply(erule conjE)

apply(simp)

apply(erule_tac x="obi" in allE)
apply(simp)
apply(rotate_tac -1)
apply(frule  mem_append_lem')
apply(simp)
apply(rule disjCI)
apply(simp)
apply(rule conjI)

apply(subgoal_tac "(ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')),{x. x = l'} \<union> snd (ob ! n))]) ! 0 = ob!0" )
apply(simp)

apply(subst nth_append)
apply(simp)

apply(rule conjI)
apply(erule conjE)

apply(subgoal_tac "((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n))]) ! Suc n) = (fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n))")

apply(simp)
apply(unfold insert_def)
apply(subgoal_tac " card ({x. x = l'} \<union> snd (ob ! n)) = Suc (card (snd (ob ! n)))")
apply(rotate_tac -1)
apply(erule ssubst)
apply(simp)+
apply(erule_tac x="fst (ob ! n)" in allE)
apply(erule_tac x="snd (ob ! n)" in allE)
apply(subst card_insert)
apply(erule impE)
apply(simp)

apply(erule conjE)
apply(subgoal_tac "finite (dom g)")
apply(subgoal_tac "snd (ob ! n) \<subseteq>  (dom g)")
apply(erule finite_subset)
apply(rule  finite_dom_fmap)
apply(simp)
apply(rule  finite_dom_fmap)
apply(subgoal_tac " snd (ob ! n)  =(snd (ob ! n) - {l'})" )
apply(rotate_tac -1)
apply(erule subst)
apply(simp)+
apply(subst nth_append)
apply(simp)

apply(rule conjI)

apply(rule allI)
apply(rule  impI)
apply(subgoal_tac "snd ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')),{x. x = l'} \<union> snd (ob ! n))]) ! i) = snd (ob!i)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac " snd ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')),{x. x = l'} \<union> snd (ob ! n))]) ! Suc n) = snd ((fst (ob ! n)(l' \<mapsto> the (g l')),{x. x = l'} \<union> snd (ob ! n)))")
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac " snd (fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n)) =  ({x. x = l'} \<union> snd (ob ! n))")
apply(rotate_tac -1)
apply(erule ssubst)
apply(erule conjE)

apply(rotate_tac 8)
apply(simp)
apply(erule_tac x="i" in allE)
apply(subgoal_tac "i \<le> n")
apply(case_tac "i=n")
apply(blast)
apply(subgoal_tac "i < n")
apply(erule impE)
apply(assumption)
apply(blast)
apply(arith)
apply(arith)
apply(simp)
apply(subst nth_append)
apply(simp)
apply(subst nth_append)
apply(simp)
apply(rule conjI)

(*finally: the main statement *)

apply(subgoal_tac "fst ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n))]) !  0)  = fst (ob!0)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac "Obj (fst (ob ! n)) T ->> Obj (fst (ob ! n)(l' \<mapsto> the (g l'))) T")
apply(subgoal_tac " (fst ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')), {x. x = l'} \<union> snd (ob ! n))]) !Suc n)) = (fst (ob ! n)(l' \<mapsto> the (g l'))) ")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rule rtranclp_trans)
apply(simp)
apply(assumption)
apply(subst nth_append)
apply(simp)

apply(erule conjE)


apply(rule obj_lem)
apply(erule_tac x ="ob ! n" in allE)
apply(simp)
apply(erule_tac x ="ob ! n" in allE)
apply(erule impE)
apply(simp)
apply(erule conjE)

(* seems like we would need drop here*)

apply(subgoal_tac " the (fst (ob ! n) l') = the (f l')")

apply(rotate_tac -1)
apply(erule ssubst)

apply(simp)+
apply(subst nth_append)
apply(simp)+

apply(subgoal_tac "card (snd ((ob @ [(fst (ob ! n)(l' \<mapsto> the (g l')), insert l' (snd (ob ! n)))]) ! Suc n)) = card ( insert l' (snd (ob ! n)))")
apply(rule impI)
apply(rule conjI)
apply(subst nth_append)+
apply(simp)
apply(erule conjE)+

apply(force)

apply(subst nth_append)
apply(subst nth_append)
apply(simp)

apply(subst nth_append)

apply(simp)
(*whew*)

apply(erule_tac x="ob!n" in allE)
apply(subgoal_tac " ob ! n mem ob")
apply(erule impE)
apply(simp)
apply(erule impE)
apply(assumption)
apply(erule conjE)
apply(erule conjE)
apply(rule impI)
apply(subgoal_tac "snd (ob ! n) \<subset>  dom f")

apply(auto) (*auto is a lot less hassle from here*)
done

lemma rtrancl_beta_obj_n: "\<lbrakk> dom f = dom g; \<forall>  l \<in> dom f. the (f l) ->>  the (g l)\<rbrakk> 
                           \<Longrightarrow> Obj f T ->> Obj g T"; 
apply (case_tac "f = empty");
apply simp;
apply(subgoal_tac " Obj empty T ->> Obj empty T")
apply(subgoal_tac "g =   empty")
apply(rotate_tac 2)
apply(erule ssubst)
apply(assumption)
apply(subst empty_dom)
apply(assumption)
apply(rule refl)
apply(rule rtranclp.intros)
apply (frule rtrancl_beta_obj_lem00);
apply assumption;
apply (rotate_tac 2);
apply (drule_tac x = "card (dom f)" in spec);
apply (erule impE);
apply (rule le_refl);
apply (erule exE);
apply (erule conjE)+;
apply (subgoal_tac "fst( ob ! 0) = f");
apply (subgoal_tac "fst (ob ! card (dom f)) = g"); 
apply simp;
prefer 2
apply (simp)
apply(erule impE)
apply(assumption)
apply(rule Ltake_eq_all)
apply(erule_tac x="ob ! card (dom f)" in allE)
apply(simp)
apply(subgoal_tac "(snd (ob ! card (dom f))) =  (dom (fst (ob ! card (dom f))))")
apply(rotate_tac -1)
apply(erule subst)
apply(erule conjE)
apply(assumption)
apply(subgoal_tac " dom (fst (ob ! card (dom f))) = dom f")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rule Ltake_eq_dom)
prefer 2
apply(assumption)
apply(erule_tac x="ob ! card (dom f)" in allE)
apply(erule impE)
apply(simp)
apply(erule conjE)
apply(erule conjE)
apply(assumption)
apply(erule_tac x="ob ! card (dom f)" in allE)
apply auto;
done

subsection {* Substitution-lemmas *}

lemma subst_eq_s [simp]: "(Var (Self k))[u,t/k] = u"
  apply (simp add: subst_Var)
  done

lemma subst_eq_p [simp]: "(Var (Param k))[u,t/k] = t"
  apply (simp add: subst_Var)
  done


lemma subst_gt_p [simp]: "i < j \<Longrightarrow> Var (Param j)[u,t/i] = Var (Param (j - 1))"
  apply (simp add: subst_Var)
  done

lemma subst_gt_s [simp]: "i < j \<Longrightarrow> Var (Self j)[u,t/i] = Var (Self (j - 1))"
  apply (simp add: subst_Var)
  done

lemma subst_lt_p [simp]: "j < i ==> Var (Param j)[u,t/i] = Var (Param j)"
  apply (simp add: subst_Var)
  done;

lemma subst_lt_s [simp]: "j < i ==> Var (Self j)[u,t/i] = Var (Self j)"
  apply (simp add: subst_Var)
  done;

lemma dB_induct: "\<lbrakk> !! n. P1 (Var n);
       !! l T. P1 (Obj l T);
       !! d1 n d2.
          \<lbrakk>P1 d1; P1 d2\<rbrakk> \<Longrightarrow> P1 (Call d1 n d2);
       !! d1 n d2.
          \<lbrakk>P1 d1; P1 d2\<rbrakk>
          \<Longrightarrow> P1 (Upd d1 n d2);
        \<And>dB. P1 dB \<Longrightarrow> P1 (Active dB);
        \<And>ActivityRef. P1 (ActRef ActivityRef);
        \<And>FutureRef. P1 (FutRef FutureRef)
  \<rbrakk> \<Longrightarrow> P1 dB" 

apply (subgoal_tac "P1 dB \<and> True \<and> True");
apply simp;
apply (erule dB.induct);
apply auto;
done;




(* Idea: "P2 f" is actually just "! x. P1 (f x)" --
   instead of a searching  for a general isomorphism between
   predicates on the range of f, i.e. f x, and f itself --
   which doesn't exist, we instantiate the induction for the case
   that we need, i.e. P2 = P1 "but lifted to fmaps" *)
(*
lemma db_induct2:"\<lbrakk> \<And>n. P1 (Var n); 
                    \<And>f. (!! x. P3 (f x)) \<Longrightarrow> P1 (Obj f);
                    \<And>d l. P1 d \<Longrightarrow> P1 (Call d l);
                    \<And>d1 l d2. \<lbrakk>P1 d1; P1 d2\<rbrakk> \<Longrightarrow> P1 (Upd d1 l d2);
                    !! x. P3 (empty x); 
                    \<And> d1 f l . \<lbrakk> l \<notin> dom f; P1 d1; (! x: dom f. P3 (f x)) \<rbrakk>
                                \<Longrightarrow> (! x: dom f. P3 ((f(l \<mapsto> d1)) x)) ;
                    !! d. P1 d \<Longrightarrow> P3 (Some d) \<rbrakk>
                  \<Longrightarrow> P1 db \<and> (! x: dom f. P3 (f x))"
*)
consts applyPropOnOption:: " (dB \<Rightarrow> bool) \<Rightarrow> dB option\<Rightarrow>bool"
primrec
f1:  "applyPropOnOption P None  = True"
f2:  "applyPropOnOption P (Some x) = P x"


lemma db_induct1:"\<lbrakk> \<And>n. P1 (Var n); 
                    \<And>f. (! x: dom f. P1 (the (f x))) \<Longrightarrow> P3 f;
                    \<And>f T. P3 f \<Longrightarrow> P1 (Obj f T);
                    \<And>d1 d2 l. \<lbrakk>P1 d1; P1 d2\<rbrakk> \<Longrightarrow> P1 (Call d1 l d2);
                    \<And>d1 l d2. \<lbrakk>P1 d1; P1 d2\<rbrakk> \<Longrightarrow> P1 (Upd d1 l d2);
                    \<And>dB. P1 dB \<Longrightarrow> P1 (Active dB);
                    \<And>ActivityRef. P1 (ActRef ActivityRef);
                    \<And>FutureRef. P1 (FutRef FutureRef) \<rbrakk>
                  \<Longrightarrow> P1 db \<and> (P3 (f::Label -~>dB ))"

apply (subgoal_tac " (\<forall>db .  P1 db) \<and> (\<forall> (f::Label -~>dB) x .(applyPropOnOption P1) (f x))");
apply (erule conjE);
apply simp;
apply (subgoal_tac " (\<forall>db (f::Label -~>dB) x . (  P1 db \<and>(applyPropOnOption P1) (f x)))");
apply simp;
apply (rule allI)+;
apply (induct_tac db and "f x");
apply force;
apply (subgoal_tac "P3 fun");
apply force;
apply (subgoal_tac " (\<forall>  xa\<in>dom fun . P1 (the (fun xa)))");
apply simp;
apply (rule ballI)
apply (drule_tac x=xa in meta_spec);
apply auto;
done;


lemma db_induct2:"\<lbrakk> \<And>n. P1 (Var n); 
                    \<And>f T. P3 f \<Longrightarrow> P1 (Obj f T);
                    \<And>d1 d2 l. \<lbrakk>P1 d1; P1 d2\<rbrakk> \<Longrightarrow> P1 (Call d1 l d2);
                    \<And>d1 l d2. \<lbrakk>P1 d1; P1 d2\<rbrakk> \<Longrightarrow> P1 (Upd d1 l d2);
                    !! x. P3 (empty); 
                    \<And> d1 f l . \<lbrakk> l \<notin> dom f; P1 d1; P3 f \<rbrakk>
                                \<Longrightarrow> ( P3 (f(l \<mapsto> d1)));
                    \<And>dB. P1 dB \<Longrightarrow> P1 (Active dB);
                    \<And>ActivityRef. P1 (ActRef ActivityRef);
                    \<And>FutureRef. P1 (FutRef FutureRef)\<rbrakk>
                  \<Longrightarrow> P1 db \<and> (P3 (f::Label -~>dB ))"

apply (subgoal_tac " (\<forall>db .  P1 db) \<and> (\<forall> (f::Label -~>dB) x .(applyPropOnOption P1) (f x))");
apply (erule conjE);
apply simp+;
apply (rule fmap_induct);
apply assumption;
apply (drule_tac x=z in meta_spec);
apply (drule_tac x=F in meta_spec)+;
apply (drule_tac x=x in meta_spec);
apply force;
apply (subgoal_tac " (\<forall>db (f::Label -~>dB) x . (  P1 db \<and>(applyPropOnOption P1) (f x)))");
apply simp;

apply (rule allI)+;
apply (induct_tac db and "f x");
apply force;
apply (subgoal_tac "P3 fun");
apply force;
apply (subgoal_tac " (\<forall>x. applyPropOnOption P1 (fun x))\<longrightarrow>P3 fun");
apply simp;
apply (rule fmap_induct);
apply simp;
apply (rule impI);
apply (rotate_tac -8);
apply (drule_tac x=z in meta_spec);
apply (rotate_tac -1);
apply (drule_tac x=F in meta_spec);
apply (rotate_tac -1);
apply (drule_tac x=xa in meta_spec);
apply (subgoal_tac "P1 z");
apply (subgoal_tac " \<forall> x. (applyPropOnOption P1 (F x))");
apply (drule mp);
apply simp;
apply force;
apply (rule allI);
apply (case_tac "xa=xb");
apply simp;
apply (subgoal_tac "F xb = None");
apply simp+;apply force;
apply (drule_tac x=xb in spec);
apply auto;
apply (drule_tac x=xa in spec);
defer;
apply (drule_tac x=xa in spec);
apply auto;
done;


lemma lift_lift_prep [rule_format]:
    "   (\<forall>i k   . i < k + 1 --> lift (lift t i) (Suc k) = lift (lift t k) i)
    &  ( \<forall>i k   .  i < k + 1  --> (lift (lift (Obj fun T) i) (Suc k) = lift (lift (Obj fun T) k) i))"

apply (rule db_induct2);
prefer 6;
apply(rule allI)+
apply (rule impI)
apply(rotate_tac -2)

apply(erule_tac x=i in allE)

apply(rotate_tac -1)
apply(erule_tac x=k in allE)
apply(erule impE)
apply(assumption)
apply simp;
apply(rule ext)
apply(case_tac "l=la")
apply simp;
apply(subgoal_tac "f l = None")
apply(subgoal_tac 
           "(if la = l then Some d1 else f la) = f la")
apply(rotate_tac -1)
apply(erule ssubst)
apply(erule fun_cong)
apply auto;
apply (case_tac n,simp+)
done

lemma lift_lift [rule_format]:
    "(\<forall>i k. i < k + 1 --> lift (lift t i) (Suc k) = lift (lift t k) i)";
apply (simp add: lift_lift_prep);
done;

lemma lift_subst_prep :
    "(\<forall>i j s t. j < i + 1 --> lift (a[s,t/j]) i = (lift a (i + 1)) [lift s i,lift t i / j])
    &(\<forall>i j s t. j < i + 1 --> lift (Obj f T [s,t/j]) i = (lift (Obj f T) (i + 1)) [lift s i,lift t i / j])";
apply (rule db_induct2);
apply (case_tac n);
apply (simp_all add: diff_Suc subst_Var lift_lift split: nat.split);

apply(rule allI)+
apply(rule impI)
apply(rule allI)+
apply(rotate_tac 2)
apply(erule_tac x=i in allE)
apply(rotate_tac 3)
apply(erule_tac x=j in allE)
apply(rule ext)

apply(erule impE)
apply(assumption)
apply(case_tac "la = l ")
apply(subgoal_tac "(if la = l then Some d1 else f la) = (Some d1)")
apply(erule ssubst)

apply(simp add: lift_lift)+
apply(rotate_tac -2)
apply(erule_tac x=s in allE)
apply(erule_tac x=t in allE)
apply(erule fun_cong)
done


lemma lift_subst [simp]:
    " j < i + 1 --> lift (t[s,u/j]) i = (lift t (i + 1)) [lift s i,lift u i / j]";
  apply (simp add: lift_subst_prep);
done;

lemma lift_subst_lt_prep:
    "(\<forall>i j s u. i < j + 1 --> lift (t[s,u/j]) i = (lift t i) [lift s i,lift u i / j + 1])
    &(\<forall>i j s u. i < j + 1 --> lift ((Obj f T)[s,u/j]) i = (lift (Obj f T) i) [lift s i,lift u i / j + 1])";
apply (rule db_induct2);
apply (case_tac n);
apply (simp_all add: subst_Var lift_lift);
apply(rule allI)+
apply(rule impI)
apply(rule allI)+
apply(rule ext)

apply(case_tac "la = l ")
apply(subgoal_tac "(if la = l then Some d1 else f la) = (Some d1)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac 2)
apply(erule_tac x=i in allE)
apply(rotate_tac -1)
apply(erule_tac x=j in allE)
apply(erule impE)
apply(assumption)
apply(rotate_tac -1)
apply(erule_tac x=s in allE)
apply(simp add: lift_lift)
apply(simp)

apply(subgoal_tac "(if la = l then (Some d1) else f la) = (f la)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac 2)
apply(erule_tac x=i in allE)
apply(rotate_tac -1)
apply(erule_tac x=j in allE)
apply(erule impE)
apply(assumption)
apply(rotate_tac -1)
apply(erule_tac x=s in allE)
apply(erule_tac x=u in allE)
apply(erule fun_cong)
apply(simp)
done


lemma lift_subst_lt:
    " i < Suc j \<Longrightarrow> lift (t[s,u/j]) i = (lift t i) [lift s i,lift u i / j + 1]";
apply (simp add: lift_subst_lt_prep);
done;

lemma subst_lift_prep :  
    "(\<forall>k s u. (lift t k)[s,u/k] = t) & (\<forall>k s u. (lift (Obj f T) k)[s,u/k] = Obj f T)";
apply (rule db_induct2);
apply (case_tac n);
apply simp_all;
apply(rule allI)+
apply(rotate_tac -1)

apply(erule_tac x=k in allE)
apply(rotate_tac -1)
apply(erule_tac x=s in allE)
apply(erule_tac x=u in allE)
apply(rule ext)

apply(case_tac " la = l")
apply(simp)
apply(subgoal_tac "(if la = l then Some d1 else f la) = (f la)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(erule fun_cong)
apply(simp)
done


lemma subst_lift [simp]: 
    "\<forall>k s u. (lift t k)[s,u/k] = t";
  apply (simp add: subst_lift_prep);
done;


lemma subst_subst_prep [rule_format]:
    "(\<forall>i j u v s s'. (i < j + 1) \<longrightarrow> 
                ((t[(lift v i),(lift s i) / Suc j])[u[v,s/j],s'[v,s/j]/i] 
                                 = t[u,s'/i][v,s/j]))
    \<and> (\<forall>i j u v s s'. (i < j + 1) \<longrightarrow> 
                ((Obj f T) [(lift v i),(lift s i) / Suc j])[(u[v,s/j]),(s'[v,s/j])/i] 
                              = Obj f T[u,s'/i][v,s/j])";
  apply (rule db_induct2);
  apply (case_tac n);
  apply (simp_all
      add: diff_Suc subst_Var lift_lift [symmetric] lift_subst_lt
      split: nat.split);

  apply(rule allI)+
  apply(rule impI)
  apply(rule allI)+
  apply(rotate_tac -2)
  apply(erule_tac x=i in allE)
  apply(rotate_tac -1)
  apply(erule_tac x=j in allE)
  apply(erule impE)
  apply(assumption)
  apply(erule_tac x=u in allE)
  apply(erule_tac x=v in allE)
  apply(erule_tac x=s in allE)
  apply(erule_tac x=s' in allE)
  apply(rule ext)
  apply(case_tac "la=l")
  apply(simp)
  apply(frule fun_cong)
  apply(simp)
done;


lemma subst_subst [rule_format]:
    " i < j + 1 \<Longrightarrow> t[lift v i,lift s' i / Suc j][u[v,s'/j],s[v,s'/j]/i] = t[u,s/i][v,s'/j]"
  apply (simp add: subst_subst_prep);
done;


(* leave out substn *)
subsection {* Equivalence proof for optimized substitution *}


subsection {* Preservation theorems *}

(* Therefore we leave them out for the moment *)


lemma insert_select [simp] : "!!f l  t. the ((f(l \<mapsto> t)) l) = t"
apply(auto)
done
lemma dom_insert [simp] : "!!f l  t. \<lbrakk> l \<in> dom f \<rbrakk> \<Longrightarrow> dom (f(l \<mapsto> t))  = dom f"
apply(auto)
done
lemma insert_select2 [simp] : "!!f l1 l2  t.\<lbrakk>l1 \<noteq> l2\<rbrakk> \<Longrightarrow>  ((f(l1 \<mapsto> t)) l2) =  (f l2)"
apply(auto)
done
lemma the_insert_select [simp] : "!!f l1 l2  t.\<lbrakk>l2 \<in> dom f; l1 \<noteq> l2\<rbrakk> \<Longrightarrow>  the ((f(l1 \<mapsto> t)) l2) =  the (f l2)"
apply(auto)
done



subsection {*Size od dB terms*}

(* this section defines the size of dB terms 
compared to size, the size of an object is the sum of the size of its fields +1 *)

constdefs fsize0 :: " (Label -~> dB) \<Rightarrow>(dB \<Rightarrow> nat)\<Rightarrow>nat"
   "fsize0 f dbs == 
        foldl (op +) 0 (map dbs  (fold (op @) (\<lambda>x. [THE y. Some y = f x]) [] (dom f)))"

primrec
   dBsize :: "dB \<Rightarrow> nat" and
   dBsize2 :: "(dB option)\<Rightarrow> nat"
where
 dBsizeVar:    "dBsize (Var i) = 0"
|  dBsizeCall:   "dBsize (Call a l b) = (dBsize a) + Suc 0 + (dBsize b) "
|  dBsizeUpd:    "dBsize (Upd a l b) = (dBsize a) + (dBsize b) + Suc 0" 
|  dBsizeObj:    "dBsize (Obj f T) = fold (op +) (\<lambda>x. dBsize2 (f x)) (Suc 0) (dom f)"
|  dBsize_Act:   "dBsize (Active a)  = dBsize a + Suc 0"
|  dBsize_ARef:  "dBsize (ActRef g)  = 0"
|  dBsize_FRef:  "dBsize (FutRef f)  = 0"
|  dBsize_Some:      "dBsize2 (Some y) = dBsize y + Suc 0"
|  dBsize_None:      "dBsize2 (None) = 0"

lemma SizeOfObjectPos: "dBsize (Obj (f::Label-~>dB) T) >0"
apply simp;
apply (subgoal_tac "finite (dom f) ")
apply (erule finite.induct);
apply (auto simp: FMap.finite_dom_fmap);
apply (case_tac "a\<in>A",auto simp: insert_absorb);
done;

end
